use anyhow::Context;
use clap::Parser;
use rcgen::{
    Certificate, CertificateParams, DistinguishedName, ExtendedKeyUsagePurpose, IsCa, KeyPair,
    KeyUsagePurpose, SanType,
};
use sleepytunny::config::Configuration;
use std::fs;
use std::path::PathBuf;

#[derive(Parser, Debug)]
#[command(
    name = "mkcert",
    author = "Ren Kararou <ren@kararou.space>",
    version = "0.1b"
)]
#[command(about = "makes simple self-signed certificates", long_about = None)]
struct Args {
    #[arg(short, long)]
    client: PathBuf,
    #[arg(short, long)]
    server: PathBuf,
}

struct Config<'a> {
    client_sni: &'a String,
    server_sni: &'a String,
    client_tls: &'a sleepytunny::config::TLS,
    server_tls: &'a sleepytunny::config::TLS,
}

fn main() -> anyhow::Result<()> {
    let args = Args::parse();
    let client_cfg = Configuration::load_config(args.client)?;
    let server_cfg = Configuration::load_config(args.server)?;
    let config = Config {
        client_sni: client_cfg
            .client()?
            .server_name()?
            .context("client config is missing server_name")?,
        server_sni: server_cfg
            .server()?
            .server_name()?
            .context("server config is missing server_name")?,
        client_tls: client_cfg.tls()?,
        server_tls: server_cfg.tls()?,
    };

    println!("Generating certificate authority...");
    let ca_cert = create_ca_cert()?;
    println!("Generating client certificate...");
    let client_cert = create_client_cert(&config)?;
    println!("Generating server certificate...");
    let server_cert = create_server_cert(&config)?;

    println!("Serializing and signing...");
    let ca_pem = ca_cert.serialize_pem()?;
    let client_pem = client_cert.serialize_pem_with_signer(&ca_cert)?;
    let client_key = client_cert.serialize_private_key_pem();
    let server_pem = server_cert.serialize_pem_with_signer(&ca_cert)?;
    let server_key = server_cert.serialize_private_key_pem();

    println!("Writing client certs...");
    fs::write(&config.client_tls.ca_file, &ca_pem)?;
    fs::write(&config.client_tls.cert_file, client_pem)?;
    fs::write(&config.client_tls.key_file, client_key)?;

    println!("Writing server certs...");
    fs::write(&config.server_tls.ca_file, &ca_pem)?;
    fs::write(&config.server_tls.cert_file, server_pem)?;
    fs::write(&config.server_tls.key_file, server_key)?;

    println!("Success");
    Ok(())
}

fn create_ca_cert() -> anyhow::Result<Certificate> {
    let key_pair = KeyPair::generate(&rcgen::PKCS_ECDSA_P256_SHA256)?;

    let mut dn = DistinguishedName::new();
    dn.push(rcgen::DnType::OrganizationName, "sleepytunny");
    dn.push(rcgen::DnType::CommonName, "rootca");

    let mut params = CertificateParams::default();
    params.is_ca = IsCa::Ca(rcgen::BasicConstraints::Unconstrained);
    params.distinguished_name = dn;
    params.key_pair = Some(key_pair);
    params.not_before = time::OffsetDateTime::now_utc();
    params.not_after = time::OffsetDateTime::now_utc() + time::Duration::days(365);
    params.key_usages = vec![KeyUsagePurpose::KeyCertSign, KeyUsagePurpose::CrlSign];

    let cert = Certificate::from_params(params)?;
    Ok(cert)
}

fn create_client_cert(_config: &Config) -> anyhow::Result<Certificate> {
    let mut dn = DistinguishedName::new();
    dn.push(rcgen::DnType::OrganizationName, "sleepytunny");
    dn.push(rcgen::DnType::CommonName, "sleepytunny client");

    let mut params = CertificateParams::default();
    params.distinguished_name = dn;
    params.not_before = time::OffsetDateTime::now_utc();
    params.not_after = time::OffsetDateTime::now_utc() + time::Duration::days(365);
    params.is_ca = IsCa::ExplicitNoCa;
    params.extended_key_usages = vec![ExtendedKeyUsagePurpose::ClientAuth];
    let cert = Certificate::from_params(params)?;
    Ok(cert)
}

fn create_server_cert(config: &Config) -> anyhow::Result<Certificate> {
    let mut dn = DistinguishedName::new();
    dn.push(rcgen::DnType::OrganizationName, "sleepytunny");
    dn.push(rcgen::DnType::CommonName, config.server_sni);

    let mut params = CertificateParams::default();
    params.distinguished_name = dn;
    params.not_before = time::OffsetDateTime::now_utc();
    params.not_after = time::OffsetDateTime::now_utc() + time::Duration::days(365);
    if config.server_sni == config.client_sni {
        params.subject_alt_names = vec![SanType::DnsName(config.server_sni.to_owned())];
    } else {
        eprintln!("WARNING: client and server have different server_name values, writing both to the server cert");
        params.subject_alt_names = vec![
            SanType::DnsName(config.server_sni.to_owned()),
            SanType::DnsName(config.client_sni.to_owned()),
        ];
    }
    params.is_ca = IsCa::ExplicitNoCa;
    params.extended_key_usages = vec![ExtendedKeyUsagePurpose::ServerAuth];
    let cert = Certificate::from_params(params)?;
    Ok(cert)
}
