use anyhow::Context;
use boring::pkey::PKey;
use boring::ssl::{SslContext, SslContextBuilder, SslMethod};
use boring::x509::store::X509StoreBuilder;
use boring::x509::X509;
use serde::{Deserialize, Serialize};
use std::fs;
use std::io::Write;
use std::net::{Ipv4Addr, SocketAddr, SocketAddrV4, ToSocketAddrs};
use std::path::Path;

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Server {
    #[serde(skip_serializing_if = "Option::is_none")]
    daemonize: Option<bool>,
    endpoint: String,
    #[serde(skip_serializing_if = "Option::is_none")]
    server_name: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    user: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    group: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    stdout: Option<String>,
    #[serde(skip_serializing_if = "Option::is_none")]
    stderr: Option<String>,
}
impl Server {
    pub fn daemonize(&self) -> anyhow::Result<bool> {
        match self.daemonize {
            Some(b) => Ok(b),
            None => Ok(true),
        }
    }
    pub fn user(&self) -> anyhow::Result<String> {
        match &self.user {
            Some(u) => Ok(u.clone()),
            None => Ok(String::from("nobody")),
        }
    }
    pub fn group(&self) -> anyhow::Result<String> {
        match &self.group {
            Some(g) => Ok(g.clone()),
            None => Ok(String::from("daemon")),
        }
    }
    pub fn stdout(&self) -> anyhow::Result<String> {
        match &self.stdout {
            Some(s) => Ok(s.clone()),
            None => Ok(String::from("/tmp/sleepyserver")),
        }
    }
    pub fn stderr(&self) -> anyhow::Result<String> {
        match &self.stderr {
            Some(s) => Ok(s.clone()),
            None => Ok(String::from("/tmp/sleepyserver.err")),
        }
    }
    /// Returns the socketaddr our endpoint should bind to
    pub fn endpoint(&self) -> anyhow::Result<SocketAddr> {
        self.endpoint
            .to_socket_addrs()?
            .next()
            .context("bad server socketaddr")
    }
    /// Returns the server name for SNI
    pub fn server_name(&self) -> anyhow::Result<Option<&String>> {
        Ok(self.server_name.as_ref())
    }
}

impl Default for Server {
    fn default() -> Self {
        Server {
            daemonize: Some(true),
            endpoint: String::from("127.0.0.1:9092"),
            server_name: Some(String::from("alphamethyl.barr0w.net")),
            user: Some(String::from("nobody")),
            group: Some(String::from("daemon")),
            stdout: Some(String::from("/tmp/sleepyserver")),
            stderr: Some(String::from("/tmp/sleepyserver.err")),
        }
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Client {
    pub endpoint: String,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub server_name: Option<String>,
}
impl Client {
    /// Returns the socketaddr that our udp socket should bind to
    pub fn local_bind_addr(&self) -> anyhow::Result<SocketAddr> {
        Ok(SocketAddrV4::new(Ipv4Addr::UNSPECIFIED, 0).into())
    }
    /// Returns the socketaddr of the server's endpoint
    pub fn endpoint(&self) -> anyhow::Result<SocketAddr> {
        self.endpoint
            .to_socket_addrs()?
            .next()
            .context("bad server socketaddr")
    }
    /// Returns the server name for SNI
    pub fn server_name(&self) -> anyhow::Result<Option<&String>> {
        Ok(self.server_name.as_ref())
    }
}

impl Default for Client {
    fn default() -> Self {
        Client {
            endpoint: String::from("alphamethyl.barr0w.net:9092"),
            server_name: Some(String::from("alphamethyl.barr0w.net")),
        }
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct TLS {
    pub ca_file: String,
    pub cert_file: String,
    pub key_file: String,
}
impl TLS {
    /// Loads in the key and cert from files specified in config.
    pub fn load(&self) -> anyhow::Result<TlsKeys> {
        TlsKeys::new(&self.ca_file, &self.cert_file, &self.key_file)
    }
}

impl Default for TLS {
    fn default() -> Self {
        TLS {
            ca_file: String::from("/etc/sleepytunny/ca.crt"),
            cert_file: String::from("/etc/sleepytunny/tls.crt"),
            key_file: String::from("/etc/sleepytunny/tls.key"),
        }
    }
}

#[derive(Debug, Clone)]
pub struct TlsKeys {
    ca: X509,
    cert: X509,
    key: PKey<boring::pkey::Private>,
}
impl TlsKeys {
    pub fn new(
        ca: impl AsRef<Path>,
        cert: impl AsRef<Path>,
        key: impl AsRef<Path>,
    ) -> anyhow::Result<Self> {
        let ca = X509::from_pem(&fs::read(ca).context("failed to read ca certificate")?)?;
        let cert = X509::from_pem(&fs::read(cert).context("failed to read tls certificate")?)?;
        let key = PKey::private_key_from_pem(&fs::read(key).context("failed to read tls key")?)?;
        Ok(Self { ca, cert, key })
    }
    pub fn into_ssl_ctx(self) -> anyhow::Result<SslContextBuilder> {
        let mut ctx = SslContext::builder(SslMethod::tls())?;
        ctx.set_servername_callback(|sslref, sslalert| {
            let servername = sslref.servername(boring::ssl::NameType::HOST_NAME);
            eprintln!("SNI callback");
            eprintln!("    servername: {servername:?}");
            eprintln!("    alert:      {sslalert:?}");
            Ok(())
        });
        let store = {
            let mut store = X509StoreBuilder::new()?;
            store.add_cert(self.ca.clone())?;
            store.build()
        };
        ctx.set_verify_cert_store(store)?;
        ctx.set_certificate(&self.cert)?;
        ctx.add_extra_chain_cert(self.ca)?;
        ctx.set_private_key(&self.key)?;
        Ok(ctx)
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Interface {
    pub address: Ipv4Addr,
    pub netmask: Ipv4Addr,
}
impl Default for Interface {
    fn default() -> Self {
        Interface {
            address: Ipv4Addr::from([192, 168, 255, 1]),
            netmask: Ipv4Addr::from([255, 255, 255, 255]),
        }
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Configuration {
    #[serde(skip_serializing_if = "Option::is_none")]
    pub server: Option<Server>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub client: Option<Client>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub tls: Option<TLS>,
    pub interface: Interface,
}
impl Configuration {
    /// Loads configuration file f and returns a struct of all parsed values
    pub fn load_config(f: impl AsRef<Path>) -> anyhow::Result<Self> {
        let config = fs::read_to_string(f);
        match config {
            Ok(c) => Ok(toml::from_str(&c)?),
            Err(_) => Ok(Self::default()),
        }
    }
    /// Generates new client configuration file f with default values
    pub fn generate_client_conf(f: &str) -> anyhow::Result<()> {
        let mut config = Self::default();
        config.server.take();
        let mut config_file = fs::File::create(f)?;
        config_file.write_all(toml::to_string(&config)?.as_bytes())?;
        Ok(())
    }
    /// Generates new server configuration file f with default values
    pub fn generate_server_conf(f: &str) -> anyhow::Result<()> {
        let mut config = Self::default();
        config.client.take();
        let mut config_file = fs::File::create(f)?;
        config_file.write_all(toml::to_string(&config)?.as_bytes())?;
        Ok(())
    }
    /// Generates new configuration file f with default values
    pub fn generate_config(f: &str) -> anyhow::Result<()> {
        let config = Self::default();
        let mut config_file = fs::File::create(f)?;
        config_file.write_all(toml::to_string(&config)?.as_bytes())?;
        Ok(())
    }
    /// Returns the server config or an appropriate error
    pub fn server(&self) -> anyhow::Result<&Server> {
        self.server
            .as_ref()
            .context("this config is not a server config")
    }
    /// Returns the client config or an appropriate error
    pub fn client(&self) -> anyhow::Result<&Client> {
        self.client
            .as_ref()
            .context("this config is not a client config")
    }
    /// Returns the TLS confix or an appropriate error
    pub fn tls(&self) -> anyhow::Result<&TLS> {
        self.tls.as_ref().context("no TLS config")
    }
}
impl Default for Configuration {
    fn default() -> Self {
        Configuration {
            server: Some(Server::default()),
            client: Some(Client::default()),
            tls: Some(TLS::default()),
            interface: Interface::default(),
        }
    }
}
